; akkulader9Vlcd.asm - Spannung messen
; ADC0 Spannung messen, LCD ausgeben
; Interner Oszi, 8MHz, Mega48
; int nach 5 Sekunden zum messen
; Beginn: 14.1.2014, Stand: 26.1.2014	
; Author: Alfred Hell  alhell@ajhe.eu
; 		funkt!
; Fuses:  01 DF E2
;
; PD6: OUT   -   Timer0, Akku laden
;
; PC0: IN    -   ADC -in
;
; PB0: OUT   -   LCD Data 0 -> PIN 11
; PB1: OUT   -   LCD Data 1 -> PIN 12
; PB2: OUT   -   LCD Data 2 -> PIN 13
; PB3: OUT   -   LCD Data 3 -> PIN 14
; PB4: OUT   -   LCD RS
; PB5: OUT   -   LCD E
;      OUT   -   LCD Kontrast
;      OUT   -   LCD plus
;      OUT   -   LCD Masse

.NOLIST
.INCLUDE "/usr/share/avra/m48def.inc"
.LIST

.EQU 	XTAL 	= 8000000

.EQU 	T1ZEIT  = 62500  ; alle 5 Sek. Ausgabe, Vorteiler=1024

.EQU 	VOLTMAX = 10166	 ; ca. 10,2 Volt = 2,15V = 442 * 23
 
.equ LCD_PORT = PORTB		; LCD port
.equ LCD_DDR  = DDRB
.equ PIN_RS   = 4
.equ PIN_E    = 5

; Definitionen:
;   Register
.DEF 	adlo   = R13
.DEF 	adhi   = R14
.DEF 	messu  = R24
.DEF 	temp1  = R16
.DEF 	temp2  = R17
.DEF 	temp3  = R18
.DEF 	temp4  = R19
.DEF 	temp5  = R20
.DEF 	temp6  = R21
.DEF 	temp7  = R22
.DEF 	temp8  = R23
.DEF 	savreg = R25
 
.CSEG
.ORG $0000
	rjmp 	main

.ORG	OC1Aaddr
	rjmp 	OC1Ahandler


OC1Ahandler:
	push 	savreg
	in 	savreg,SREG
	sbi 	GPIOR0, 0
	out 	SREG,savreg
	pop 	savreg
	reti

main:
	ldi 	temp5,LOW(RAMEND) 	; Richte den Stapel ein
	out 	SPL,temp5
	ldi 	temp5,HIGH(RAMEND) 	; Richte den Stapel ein
	out 	SPH,temp5

	ldi	temp5,0b01000000
	out 	DDRD,temp5		; Port D6 Ausgang
	ldi 	temp5,16
	out 	OCR0A,temp5		; 112Hz, 6% Auszeit
	ldi 	temp5,(1<<COM0A1)|(1<<WGM00)|(1<<WGM01) ; Timer0 Fast PWM 
	out 	TCCR0A,temp5				; clear at OCR0A, Set at TOP
	ldi 	temp5,1<<CS02
	out 	TCCR0B,temp5		;Timer0 einschalten, Vorteiler 256

	ldi 	temp5,LOW(T1ZEIT)   	; Timer1 Messzeitpunkt einstellen
	ldi 	temp6,HIGH(T1ZEIT)	; ca. alle 5 Sekunden
	sts 	OCR1AH,temp6		; zuerst Highbyte schreiben!
	sts 	OCR1AL,temp5
	ldi 	temp5,(1<<wGM12)|(1<<CS12)|(1<<CS10)
	sts 	TCCR1B,temp5		;Timer1 einschalten, Vorteiler 1024, CTC Modus
	ldi 	temp5,1<<OCIE1A 	; Timer1 interupt ein
	sts 	TIMSK1,temp5

				; setze MUX auf Kanal 0, Rechtsjustieren, AREF auf AVCC
	ldi 	temp5, 1<<REFS0 	; ADMUX Kanal 0, AREF auf AVCC
	sts 	ADMUX,temp5
				; AD-Wandler einschalten, Teilerrate = 64 = 125kHz
	ldi 	temp5,(1<<ADEN)|(1<<ADPS2)|(1<<ADPS1)
	sts 	ADCSRA,temp5

        rcall   lcd_init
        rcall   lcd_clear
	rcall	lcd_home
	ldi 	ZL, LOW(text5*2)         ; Adresse des Strings in den
	ldi	ZH, HIGH(text5*2)        ; Z-Pointer laden
	rcall 	lcd_flash_string      ; Unterprogramm gibt String an LCD
	rcall	lcd_zeile2
	ldi	ZL, LOW(text6*2)        
	ldi	ZH, HIGH(text6*2)       
	rcall	lcd_flash_string    

	clr 	XL			; zaehlregister leeren
	clr 	YL
	clr 	YH

	ldi	temp5, 0xFF		; vor 1. Messung warten...
lay1s: rcall	delay5ms
	dec	temp5
	brne	lay1s

	sei				; global interupt enable

no_end:   ; unendliche Schleife
	sbic 	GPIOR0, 0
	rjmp 	messen
	rjmp 	no_end

messen: 
	cbi	GPIOR0, 0
	clr 	temp5
	out 	TCCR0B,temp5	;Timer0 auschalten
	out 	TCCR0A,temp5	
	adiw	YH:YL,8		; Sekundenzaehler erhoehen
	cbi	PORTD,PD6	; laden abschalten

	rcall	delay5ms	; Pause vor Messung
	clr     temp1
    	clr     temp2
    	clr     temp3
 	clr     temp4
	clr     temp5
    	clr     temp6
    	clr     temp7
 	clr     temp8
	ldi	messu, 23	; neuen ADC-Wert lesen  (Schleife - 23 mal)
sample_adc: 			; messen start
	ldi 	temp5,(1<<ADSC)|(1<<ADEN)|(1<<ADPS2)|(1<<ADPS1)
	sts 	ADCSRA,temp5

wait_adc:
    	lds	temp5,ADCSRA	
	sbrc    temp5,6		    ; wenn der ADC fertig ist, wird dieses Bit gelöscht
    	rjmp    wait_adc
 
				; ADC einlesen:
	lds      adlo, ADCL         ; immer zuerst LOW Byte lesen
    	lds      adhi, ADCH        ; danach das mittlerweile gesperrte High Byte
 
; alle 256 ADC-Werte addieren
; dazu wird mit den Registern temp4, temp3 und temp2 ein
; 24-Bit breites Akkumulationsregister gebildet, in dem
; die 10 Bit Werte aus adhigh, adlow aufsummiert werden
 
    	add     temp7, adlo        ; addieren
    	adc     temp8, adhi       ; addieren über Carry
    	dec     messu             ; Schleifenzähler MINUS 1
    	brne    sample_adc          ; wenn noch keine 256 ADC Werte -> nächsten Wert einlesen
 
	mov 	R3,temp8		; in die passenden R für Wandlung
	mov 	R2,temp7

	ldi	temp5,LOW(VOLTMAX)		; ab 10,2 Volt erhaltungsladen
	ldi	temp6,HIGH(VOLTMAX)
	cp	R2,temp5		
	cpc	R3,temp6		; vergleichen
	brcs	weite			; wenn R2 kleiner, weiterladen

	rcall   lcd_clear
	rcall	lcd_home
	ldi 	ZL, LOW(text4*2)         ; Adresse des Strings in den
    	ldi     ZH, HIGH(text4*2)            ; Z-Pointer laden
	rcall	lcd_flash_string	     ; Unterprogramm gibt String aus der

	ldi 	temp5,(1<<COM0A1)|(1<<COM0A0)|(1<<WGM00)|(1<<WGM01) ; Timer0 Fast PWM 
	out 	TCCR0A,temp5				; Set at OCR0A, clear at Top
	rjmp	weiter

weite:
	rcall   lcd_clear
	rcall	lcd_home
	ldi 	ZL, LOW(text*2)         ; Adresse des Strings in den
	ldi 	ZH, HIGH(text*2)        ; Z-Pointer laden
	rcall 	lcd_flash_string      ; Unterprogramm gibt String aus der
	
	ldi 	temp5,(1<<COM0A1)|(1<<WGM00)|(1<<WGM01) ; Timer0 Fast PWM 
	out 	TCCR0A,temp5				; Clear at OCR0A, Set at Top
	
weiter:					;in ASCII umwandeln
	ldi 	temp5,1<<CS02
	out 	TCCR0B,temp5		;Timer0 wieder einschalten Vorteiler 256

	mov	R1, R2
	Mov	R2, R3
	rcall 	fpconv10a 		; Rufe die Umwandlungsroutine

	mov	temp1,R5		    ; Spannung ausgeben
        rcall 	lcd_data
	mov	temp1,R6
        rcall 	lcd_data
	ldi	temp1,'.'
        rcall 	lcd_data
	mov	temp1,R7
        rcall 	lcd_data
	mov	temp1,R8
        rcall 	lcd_data

    	ldi     ZL, LOW(text1*2)             ; Adresse des Strings in den
    	ldi     ZH, HIGH(text1*2)            ; Z-Pointer laden
	rcall	lcd_flash_string      ; Unterprogramm gibt String aus der

	rcall 	lcd_zeile2
    	ldi     ZL, LOW(text2*2)             ; Adresse des Strings in den
    	ldi     ZH, HIGH(text2*2)            ; Z-Pointer laden
	rcall	lcd_flash_string      ; Unterprogramm gibt String aus der

	mov 	R2,YH			; in die passenden R für Wandlung
	mov 	R1,YL     		; Sekunden umrechnen
	rcall 	fpconv10a 		; Rufe die Umwandlungsroutine

	mov	temp1,R5		; Zeit ausgeben
        rcall 	lcd_data
	mov	temp1,R6
        rcall 	lcd_data
	mov	temp1,R7
        rcall 	lcd_data
	mov	temp1,R8
        rcall 	lcd_data
	mov	temp1,R9
        rcall 	lcd_data

    	ldi     ZL, LOW(text3*2)             ; Adresse des Strings in den
    	ldi     ZH, HIGH(text3*2)            ; Z-Pointer laden
	rcall	lcd_flash_string      ; Unterprogramm gibt String aus der

    	rjmp 	no_end


; Wandle das Wort in R2:R1 in einen ASCII-Text in R5:R6:R7:R8:R9:R10
;
; +---+---+
; + R2| R1| Eingangswert 0..5.000
; +---+---+
; +---+---+
; | R4| R3| Dezimalteiler
; +---+---+
; *---*---+---+---+---+---+
; | R5| R6| R7| R8| R9|R10| Ergebnistext (fuer Eingangswert 5,000)
; |'0'|'0'|'0'|'0'|'0'|$00| mit Null-Abschluss
; +---+---+---+---+---+---+
fpconv10a:
	ldi 	temp5,HIGH(10000) 	; Setze Dezimalteiler auf 10.000
	mov 	R4,temp5
	ldi 	temp5,LOW(10000)
	mov 	R3,temp5
	rcall 	fpconv10d 		; Hole ASCII-Ziffer durch wiederholtes Abziehen
	mov 	R5,temp5 		; Setze Tausender Zeichen
	clr 	R4 			; Setze Dezimalteiler auf 100
	ldi 	temp5,HIGH(1000) 	; Setze Dezimalteiler auf 1.000
	mov 	R4,temp5
	ldi 	temp5,LOW(1000)
	mov 	R3,temp5
	rcall 	fpconv10d 		; Hole ASCII-Ziffer durch wiederholtes Abziehen
	mov 	R6,temp5 		; Setze Tausender Zeichen
	clr 	R4 			; Setze Dezimalteiler auf 100
	ldi 	temp5,100
	mov 	R3,temp5
	rcall 	fpconv10d 		; Hole die naechste Ziffer
	mov 	R7,temp5 		; Setze Hunderter Zeichen
	ldi 	temp5,10 		; Setze Dezimalteiler auf 10
	mov 	R3,temp5
	rcall 	fpconv10d 		; Hole die naechste Ziffer
	mov 	R8,temp5 		; Setze Zehner Zeichen
	ldi 	temp5,'0' 		; Wandle Rest in ASCII-Ziffer um
	add 	temp5,R1
	mov 	R9,temp5 		; Setze Einer Zeichen
	ret
;
; Wandle Binaerwort in R2:R1 in eine Dezimalziffer durch fortgesetztes
; Abziehen des Dezimalteilers in R4:R3 (1000, 100, 10)
;
fpconv10d:
	ldi 	temp5,'0' 	; Beginne mit ASCII-0
fpconv10d1:
	cp 	R1,R3 		; Vergleiche Wort mit Teiler
	cpc 	R2,R4
	brcc 	fpconv10d2 	; Carry nicht gesetzt, subtrahiere Teiler
	ret 			; fertig	
fpconv10d2:
	sub 	R1,R3   	; Subtrahiere Teilerwert
	sbc 	R2,R4
	inc 	temp5 		; Ziffer um eins erhoehen
	rjmp 	fpconv10d1 	; und noch einmal von vorne
; Ende der Fliesskomma-Umwandlungsroutinen
;

 ;sendet ein Datenbyte an das LCD
lcd_data:
        push  temp2
        push  temp3
        mov   temp2, temp1            ; "Sicherungskopie" fuer
                                        ; die Uebertragung des 2.Nibbles
        swap  temp1                   ; Vertauschen
        andi  temp1, 0b00001111       ; oberes Nibble auf Null setzen
        sbr   temp1, 1<<PIN_RS        ; entspricht 0b00010000
        in    temp3, LCD_PORT
        andi  temp3, 0x80
        or    temp1, temp3
        out   LCD_PORT, temp1         ; ausgeben
        rcall lcd_enable              ; Enable-Routine aufrufen
                                         ; 2. Nibble, kein swap da es schon
                                         ; an der richtigen stelle ist
        andi  temp2, 0b00001111       ; obere Haelfte auf Null setzen 
        sbr   temp2, 1<<PIN_RS        ; entspricht 0b00010000
        or    temp2, temp3
        out   LCD_PORT, temp2         ; ausgeben
        rcall lcd_enable              ; Enable-Routine aufrufen
        rcall delay50us               ; Delay-Routine aufrufen

        pop   temp3
        pop   temp2
        ret                           ; zurueck zum Hauptprogramm
 
 ; sendet einen Befehl an das LCD
lcd_command:                            ; wie lcd_data, nur ohne RS zu setzen
        push  temp2
        push  temp3

        mov   temp2, temp1
        swap  temp1
        andi  temp1, 0b00001111
        in    temp3, LCD_PORT
        andi  temp3, 0x80
        or    temp1, temp3
        out   LCD_PORT, temp1
        rcall lcd_enable
        andi  temp2, 0b00001111
        or    temp2, temp3
        out   LCD_PORT, temp2
        rcall lcd_enable
        rcall delay50us
 
        pop   temp3
        pop   temp2
        ret
 
 ; erzeugt den Enable-Puls
lcd_enable:
        sbi 	LCD_PORT, PIN_E          ; Enable high
        nop                          ; 3 Taktzyklen warten
        nop
        nop
        cbi   	LCD_PORT, PIN_E          ; Enable wieder low
        ret                          ; Und wieder zurueck                     
 
 ; Pause nach jeder Uebertragung
delay50us:                              ; 50us Pause
        ldi  	temp1, 133  		;( XTAL * 50 / 3 ) / 1000000
delay50us_:
        dec  	temp1
        brne 	delay50us_
        ret                          ; wieder zur�ck
 
 ; Laengere Pause fuer manche Befehle
delay5ms:                               ; 5ms Pause
        ldi  	temp1, 52
WGLOOP0:
	ldi  	temp2, 0xFF
WGLOOP1:
	dec  	temp2
        brne 	WGLOOP1
        dec  	temp1
        brne 	WGLOOP0
        ret                          ; wieder zurueck
 
 ; Initialisierung: muss ganz am Anfang des Programms aufgerufen werden
lcd_init:
        push  	temp1
        in    	temp1, LCD_DDR
        ori   	temp1, (1<<PIN_E) | (1<<PIN_RS) | 0x0F
        out   	LCD_DDR, temp1
        ldi   	temp3,6
powerupwait:
        rcall 	delay5ms
        dec   	temp3
        brne  	powerupwait
        ldi   	temp1,    0b00000011   ; muss 3mal hintereinander gesendet
        out   	LCD_PORT, temp1        ; werden zur Initialisierung
        rcall 	lcd_enable             ; 1
        rcall 	delay5ms
        rcall 	lcd_enable             ; 2
        rcall 	delay5ms
        rcall 	lcd_enable             ; und 3!
        rcall 	delay5ms
        ldi   	temp1,    0b00000010   ; 4bit-Modus einstellen
        out   	LCD_PORT, temp1
        rcall 	lcd_enable
        rcall 	delay5ms
        ldi   	temp1,    0b00101000   ; 4 Bot, 2 Zeilen
        rcall 	lcd_command
        ldi   	temp1,    0b00001100   ; Display on, Cursor off
        rcall 	lcd_command
        ldi   	temp1,    0b00000100   ; endlich fertig
        rcall 	lcd_command
        pop   	temp1
        ret
 
 ; Sendet den Befehl zur Loeschung des Displays
lcd_clear:
        push  	temp1
        ldi   	temp1,    0b00000001   ; Display loeschen
        rcall 	lcd_command
        rcall 	delay5ms
        pop   	temp1
        ret

 ; Cursor Home
lcd_home:
        push  	temp1
        ldi   	temp1,    0b00000010   ; Cursor Home
        rcall 	lcd_command
        rcall 	delay5ms
        pop   	temp1
        ret

; cursor 2. Zeile
lcd_zeile2:
        push  	temp1
        ldi   	temp1,    0b11000000   ; Cursor Zeil2
        rcall 	lcd_command
        rcall 	delay5ms
        pop   	temp1
        ret

 ; Einen konstanten Text aus dem Flash Speicher
 ; ausgeben. Der Text wird mit einer 0 beendet
lcd_flash_string:
        push  	temp1

lcd_flash_string_1:
        lpm   	temp1, Z+
        cpi   	temp1, 0
        breq  	lcd_flash_string_2
        rcall 	lcd_data
        rjmp  	lcd_flash_string_1
lcd_flash_string_2:
        pop   	temp1
        ret

text:
       .db "Volt:  ",0                ; Stringkonstante, durch eine 0
                                       ; abgeschlossen  
text1:
       .db " V ",0                ; Stringkonstante, durch eine 0
                                       ; abgeschlossen  
text2:
       .db "Zeit:  ",0                ; Stringkonstante, durch eine 0

Text3:                                       ; abgeschlossen  
       .db " Sek ",0                ; Stringkonstante, durch eine 0
                                       ; abgeschlossen  
text4:
       .db "ELaden:",0                ; Stringkonstante, durch eine 0
                                       ; abgeschlossen  
text5:
       .db "9Volt Akkulader",0                ; Stringkonstante, durch eine 0
                                       ; abgeschlossen  
text6:
       .db "(c) Alfred Hell",0                ; Stringkonstante, durch eine 0
                                       ; abgeschlossen  

; ende
